package main

import (
	"bufio"
	"bytes"
	"crypto/tls"
	"crypto/x509"
	"fmt"
	"log"
	"net"
	"net/url"
	"os"
	"path"
	"regexp"
	"strconv"
	"strings"
	"sync"
	"time"
	"unicode/utf8"
)

const (
	readTimeout = 30 * time.Second

	urlMaxLength = 1024

	geminiType = "text/gemini; charset=utf-8"

	logTimeFormat = "2006-01-02 15:04:05"
)

const (
	statusInput          = 10
	statusSensitiveInput = 11

	statusSuccess = 20

	statusRedirectTemporary = 30
	statusRedirectPermanent = 31

	statusTemporaryFailure = 40
	statusUnavailable      = 41
	statusCGIError         = 42
	statusProxyError       = 43

	statusPermanentFailure    = 50
	statusNotFound            = 51
	statusGone                = 52
	statusProxyRequestRefused = 53
	statusBadRequest          = 59
)

var slashesRegexp = regexp.MustCompile(`[^\\]\/`)

var newLine = "\r\n"

var logLock sync.Mutex

func writeHeader(c net.Conn, code int, meta string) int {
	fmt.Fprintf(c, "%d %s%s", code, meta, newLine)
	return code
}

func writeStatus(c net.Conn, code int) int {
	var meta string
	switch code {
	case statusTemporaryFailure:
		meta = "Temporary failure"
	case statusProxyError:
		meta = "Proxy error"
	case statusBadRequest:
		meta = "Bad request"
	case statusNotFound:
		meta = "Not found"
	case statusProxyRequestRefused:
		meta = "Proxy request refused"
	}
	writeHeader(c, code, meta)
	return code
}

func writeSuccess(c net.Conn, serve *pathConfig, contentType string, size int64) int {
	meta := contentType
	if serve.Type != "" {
		meta = serve.Type
	}

	if !config.DisableSize && size >= 0 {
		meta += fmt.Sprintf("; size=%d", size)
	}

	if serve.cache != cacheUnset {
		meta += fmt.Sprintf("; cache=%d", serve.cache)
	}

	writeHeader(c, statusSuccess, meta)
	return statusSuccess
}

func scanCRLF(data []byte, atEOF bool) (advance int, token []byte, err error) {
	if atEOF && len(data) == 0 {
		return 0, nil, nil
	}
	if i := bytes.IndexByte(data, '\r'); i >= 0 {
		// We have a full newline-terminated line.
		return i + 1, data[0:i], nil
	}
	// If we're at EOF, we have a final, non-terminated line. Return it.
	if atEOF {
		return len(data), data, nil
	}
	// Request more data.
	return 0, nil, nil
}

func replaceWithUserInput(command []string, request *url.URL) []string {
	newCommand := make([]string, len(command))
	copy(newCommand, command)
	for i, piece := range newCommand {
		if strings.Contains(piece, "$USERINPUT") {
			requestQuery, err := url.QueryUnescape(request.RawQuery)
			if err == nil {
				newCommand[i] = strings.ReplaceAll(piece, "$USERINPUT", requestQuery)
			}
		}
	}
	return newCommand
}

func servePath(c *tls.Conn, request *url.URL, serve *pathConfig) (int, int64) {
	resolvedPath := request.Path
	requestSplit := strings.Split(request.Path, "/")
	pathSlashes := len(slashesRegexp.FindAllStringIndex(serve.Path, -1))
	if len(serve.Path) > 0 {
		if serve.Path[0] == '/' {
			pathSlashes++ // Regexp does not match starting slash
		}
		if serve.Path[len(serve.Path)-1] != '/' {
			pathSlashes++
		}
	}
	if len(requestSplit) >= pathSlashes {
		resolvedPath = strings.Join(requestSplit[pathSlashes:], "/")
	}

	if !serve.HiddenFiles {
		for _, piece := range requestSplit {
			if len(piece) > 0 && piece[0] == '.' {
				return writeStatus(c, statusTemporaryFailure), -1
			}
		}
	}

	var filePath string
	if serve.Root != "" {
		root := serve.Root
		if root[len(root)-1] != '/' {
			root += "/"
		}

		if !serve.SymLinks {
			for i := range requestSplit[pathSlashes:] {
				info, err := os.Lstat(path.Join(root, strings.Join(requestSplit[pathSlashes:pathSlashes+i+1], "/")))
				if err != nil || info.Mode()&os.ModeSymlink == os.ModeSymlink {
					return writeStatus(c, statusTemporaryFailure), -1
				}
			}
		}

		filePath = path.Join(root, resolvedPath)
	}

	if serve.Proxy != "" {
		return serveProxy(c, request, serve.Proxy), -1
	} else if serve.FastCGI != "" {
		if filePath == "" {
			return writeStatus(c, statusNotFound), -1
		}

		contentType := geminiType
		if serve.Type != "" {
			contentType = serve.Type
		}
		writeSuccess(c, serve, contentType, -1)

		serveFastCGI(c, config.fcgiPools[serve.FastCGI], request, filePath)
		return statusSuccess, -1
	} else if serve.cmd != nil {
		requireInput := serve.Input != "" || serve.SensitiveInput != ""
		if requireInput {
			newCommand := replaceWithUserInput(serve.cmd, request)
			if newCommand != nil {
				return serveCommand(c, serve, request, newCommand)
			}
		}
		return serveCommand(c, serve, request, serve.cmd)
	}

	if filePath == "" {
		return writeStatus(c, statusNotFound), -1
	}

	fi, err := os.Stat(filePath)
	if err != nil {
		return writeStatus(c, statusNotFound), -1
	}

	mode := fi.Mode()
	hasTrailingSlash := len(request.Path) > 0 && request.Path[len(request.Path)-1] == '/'
	if mode.IsDir() {
		if !hasTrailingSlash {
			return writeHeader(c, statusRedirectPermanent, request.String()+"/"), -1
		}

		_, err := os.Stat(path.Join(filePath, "index.gmi"))
		if err != nil {
			_, err := os.Stat(path.Join(filePath, "index.gemini"))
			if err != nil {
				if serve.List {
					return serveDirList(c, serve, request, filePath), -1
				}
				return writeStatus(c, statusNotFound), -1
			}
			filePath = path.Join(filePath, "index.gemini")
		} else {
			filePath = path.Join(filePath, "index.gmi")
		}
	} else if hasTrailingSlash && len(request.Path) > 1 {
		r := request.String()
		return writeHeader(c, statusRedirectPermanent, r[:len(r)-1]), -1
	}

	serveFile(c, serve, filePath)
	return statusSuccess, fi.Size()
}

func handleRequest(c *tls.Conn, request *url.URL, requestData string) (int, int64, string) {
	if request.Path == "" {
		// Redirect to /
		return writeHeader(c, statusRedirectPermanent, requestData+"/"), -1, ""
	}

	pathBytes := []byte(request.Path)
	strippedPath := request.Path
	if strippedPath[0] == '/' {
		strippedPath = strippedPath[1:]
	}

	var matchedHost bool
	requestHostname := request.Hostname()
	for hostname := range config.Hosts {
		if requestHostname != hostname {
			continue
		}
		matchedHost = true

		for _, serve := range config.Hosts[hostname].Paths {
			matchedRegexp := serve.r != nil && serve.r.Match(pathBytes)
			matchedPrefix := serve.r == nil && strings.HasPrefix(request.Path, serve.Path)
			if !matchedRegexp && !matchedPrefix {
				continue
			}

			requireInput := serve.Input != "" || serve.SensitiveInput != ""
			if request.RawQuery == "" && requireInput {
				if serve.Input != "" {
					return writeHeader(c, statusInput, serve.Input), -1, ""
				} else if serve.SensitiveInput != "" {
					return writeHeader(c, statusSensitiveInput, serve.SensitiveInput), -1, ""
				}
			}

			if matchedRegexp || matchedPrefix {
				status, size := servePath(c, request, serve)
				return status, size, serve.Log
			}
		}
		break
	}

	if matchedHost {
		return writeStatus(c, statusNotFound), -1, ""
	}
	return writeStatus(c, statusProxyRequestRefused), -1, ""
}

func handleConn(c *tls.Conn) {
	t := time.Now()
	var request *url.URL
	var logPath string
	status := 0
	size := int64(-1)

	defer func() {
		if !verbose && logPath == "" {
			return
		}

		entry := logEntry(request, status, size, time.Since(t))

		if verbose {
			log.Println(string(entry))
		}

		if logPath == "" {
			return
		}

		logLock.Lock()
		defer logLock.Unlock()

		f, err := os.OpenFile(logPath, os.O_APPEND|os.O_WRONLY|os.O_CREATE, 0600)
		if err != nil {
			log.Printf("ERROR: Failed to open log file at %s: %s", logPath, err)
			return
		}
		defer f.Close()

		if _, err = f.Write(entry); err != nil {
			log.Printf("ERROR: Failed to write to log file at %s: %s", logPath, err)
			return
		}
		f.Write([]byte("\n"))
	}()

	defer c.Close()
	c.SetReadDeadline(time.Now().Add(readTimeout))

	var requestData string
	scanner := bufio.NewScanner(c)
	if !config.SaneEOL {
		scanner.Split(scanCRLF)
	}
	if scanner.Scan() {
		requestData = scanner.Text()
	}
	if err := scanner.Err(); err != nil {
		status = writeStatus(c, statusBadRequest)
		return
	}

	state := c.ConnectionState()
	certs := state.PeerCertificates
	var clientCertKeys [][]byte
	for _, cert := range certs {
		pubKey, err := x509.MarshalPKIXPublicKey(cert.PublicKey)
		if err != nil {
			continue
		}
		clientCertKeys = append(clientCertKeys, pubKey)
	}

	if len(requestData) > urlMaxLength || !utf8.ValidString(requestData) {
		status = writeStatus(c, statusBadRequest)
		return
	}

	var err error
	request, err = url.Parse(requestData)
	if err != nil {
		status = writeStatus(c, statusBadRequest)
		return
	}

	requestHostname := request.Hostname()
	if requestHostname == "" || strings.ContainsRune(requestHostname, ' ') {
		status = writeStatus(c, statusBadRequest)
		return
	}

	var requestPort int
	if request.Port() != "" {
		requestPort, err = strconv.Atoi(request.Port())
		if err != nil {
			requestPort = 0
		}
	}

	if request.Scheme == "" {
		request.Scheme = "gemini"
	}
	if request.Scheme != "gemini" || (requestPort > 0 && requestPort != config.port) {
		status = writeStatus(c, statusProxyRequestRefused)
		return
	}

	status, size, logPath = handleRequest(c, request, requestData)
}

func logEntry(request *url.URL, status int, size int64, elapsed time.Duration) []byte {
	hostFormatted := "-"
	pathFormatted := "-"
	sizeFormatted := "-"
	if request != nil {
		if request.Path != "" {
			pathFormatted = request.Path
		}
		if request.Hostname() != "" {
			hostFormatted = request.Hostname()
			if request.Port() != "" {
				hostFormatted += ":" + request.Port()
			} else {
				hostFormatted += ":1965"
			}
		}
	}
	if size >= 0 {
		sizeFormatted = strconv.FormatInt(size, 10)
	}
	return []byte(fmt.Sprintf(`%s - - - [%s] "GET %s Gemini" %d %s %.4f`, hostFormatted, time.Now().Format(logTimeFormat), pathFormatted, status, sizeFormatted, elapsed.Seconds()))
}

func handleListener(l net.Listener) {
	for {
		conn, err := l.Accept()
		if err != nil {
			log.Fatal(err)
		}

		go handleConn(conn.(*tls.Conn))
	}
}

func getCertificate(info *tls.ClientHelloInfo) (*tls.Certificate, error) {
	host := config.Hosts[info.ServerName]
	if host != nil {
		return host.cert, nil
	}
	for _, host := range config.Hosts {
		return host.cert, nil
	}
	return nil, nil
}

func listen(address string) {
	tlsConfig := &tls.Config{
		ClientAuth:     tls.RequestClientCert,
		GetCertificate: getCertificate,
	}

	listener, err := tls.Listen("tcp", address, tlsConfig)
	if err != nil {
		log.Fatalf("failed to listen on %s: %s", address, err)
	}

	handleListener(listener)
}
